@@ -1,13 +1,10 @@
-from collections import Counter
 from itertools import chain
 from typing import Union, List
-import re
 from lazy import lazy
 
 import category_encoders as ce
 import numpy as np
 import pandas as pd
-from pandas._libs.tslibs.parsing import guess_datetime_format
 import tensorflow as tf
 import tensorflow.keras.backend as K
 from scipy.stats import shapiro
@@ -20,10 +17,15 @@ from tensorflow.keras.layers import (
     LSTM,
     Layer,
     RepeatVector,
-    TimeDistributed
+    TimeDistributed,
 )
 
-from syngen.ml.utils import slugify_parameters, inverse_dict
+from syngen.ml.utils import (
+    slugify_parameters,
+    inverse_dict,
+    datetime_to_timestamp,
+    timestamp_to_datetime
+)
 
 
 class BaseFeature:
@@ -33,20 +35,21 @@ class BaseFeature:
     What is more, each feature class contains modules for the neural network (NN), including
     corresponding input, encoder, decoder, and loss
     """
+
     def __init__(self, name):
         self.name: str = self._reset_name(name=name)
         self.original_name: str = name
         self.weight: float = 1.0
 
     @staticmethod
-    @slugify_parameters()
+    @slugify_parameters(turn_on=False)
     def _reset_name(name):
         """
         Slugify the attribute 'name' of the instance
         """
         return name
 
-    def fit(self, data: pd.DataFrame):
+    def fit(self, data: pd.DataFrame, **kwargs):
         """
         Fit scalars, one-hot encoders, and mappers in the data
         """
@@ -99,16 +102,19 @@ class BinaryFeature(BaseFeature):
     """
     A class to process binary features, i.e. features containing only two unique values
     """
+
     def __init__(self, name: str):
         super().__init__(name=name)
+        self.feature_type = "categorical"
 
-    def fit(self, data: pd.DataFrame):
+    def fit(self, data: pd.DataFrame, **kwargs):
         self.mapping = {k: n for n, k in enumerate(np.unique(data))}
         self.inverse_mapping = inverse_dict(self.mapping)
         self.inverse_vectorizer = np.vectorize(self.inverse_mapping.get)
         self.input_dimension = data.shape[1]
 
-    def transform(self, data: pd.DataFrame) -> List:
+    def transform(self, data: pd.DataFrame):
+        data = data.astype("object").replace(self.mapping)
         data = data.replace(self.mapping)
         return data.astype("float64")
 
@@ -145,6 +151,7 @@ class ContinuousFeature(BaseFeature):
     """
     A class to process the continuous numeric features, including floats and integers
     """
+
     def __init__(
         self,
         name: str,
@@ -171,8 +178,9 @@ class ContinuousFeature(BaseFeature):
         self.decoder_layers = decoder_layers
         self.weight_randomizer = weight_randomizer
         self.column_type = column_type
+        self.feature_type = "numeric"
 
-    def fit(self, data: pd.DataFrame):
+    def fit(self, data: pd.DataFrame, **kwargs):
         self.is_positive = (data >= 0).sum().item() >= len(data) * 0.99
         normality = shapiro(data.sample(n=min(len(data), 500))).pvalue
         self.scaler = StandardScaler() if normality >= 0.05 else MinMaxScaler()
@@ -184,7 +192,9 @@ class ContinuousFeature(BaseFeature):
 
     def inverse_transform(self, data: np.ndarray) -> np.ndarray:
         reverse_transformed = self.scaler.inverse_transform(data)
-        reverse_transformed = np.abs(reverse_transformed) if self.is_positive else reverse_transformed
+        reverse_transformed = (
+            np.abs(reverse_transformed) if self.is_positive else reverse_transformed
+        )
         return (
             reverse_transformed
             if self.column_type is float
@@ -212,9 +222,7 @@ class ContinuousFeature(BaseFeature):
                 decoder_layers.append(Layer)
 
         decoder_layers.append(
-            Dense(
-                self.input_dimension, activation="linear", name="%s_linear" % self.name
-            )
+            Dense(self.input_dimension, activation="linear", name="%s_linear" % self.name)
         )
 
         return decoder_layers
@@ -248,13 +256,13 @@ class CategoricalFeature(BaseFeature):
     """
     A class to process categorical values, i.e. values with 2 < unique_values < 50
     """
+
     def __init__(
         self,
         name: str,
         decoder_layers: Union[None, tuple, int] = (60,),
         weight_randomizer: Union[None, bool, tuple] = None,
     ):
-
         if decoder_layers is None:
             decoder_layers = ()
         elif isinstance(decoder_layers, int):
@@ -270,18 +278,19 @@ class CategoricalFeature(BaseFeature):
             weight_randomizer = (0, 1)
 
         super().__init__(name="_".join(name.split()))
-        self.one_hot_encoder = ce.OneHotEncoder(
-            return_df=False, handle_unknown="ignore"
-        )
+        self.one_hot_encoder = ce.OneHotEncoder(return_df=False, handle_unknown="ignore")
         self.decoder = None
         self.decoder_layers = decoder_layers
         self.weight_randomizer = weight_randomizer
+        self.feature_type = "categorical"
 
-    def fit(self, data: pd.DataFrame):
+    def fit(self, data: pd.DataFrame, **kwargs):
         data = data.astype(object)
         self.one_hot_encoder.fit(data)
         self.mapping = {
-            k: v for k, v in self.one_hot_encoder.category_mapping[0]["mapping"].items()
+            k: v
+            for k, v
+            in self.one_hot_encoder.ordinal_encoder.category_mapping[0]["mapping"].items()
         }
         self.inverse_mapping = inverse_dict(self.mapping)
         self.inverse_vectorizer = np.vectorize(self.inverse_mapping.get)
@@ -304,9 +313,7 @@ class CategoricalFeature(BaseFeature):
 
     @lazy
     def input(self) -> tf.Tensor:
-        self.idx_input = Input(
-            shape=(self.input_dimension,), name="input_%s" % self.name
-        )
+        self.idx_input = Input(shape=(self.input_dimension,), name="input_%s" % self.name)
 
         return self.idx_input
 
@@ -358,15 +365,14 @@ class CategoricalFeature(BaseFeature):
         high = self.weight_randomizer[1]
         random_weight = K.random_uniform_variable(shape=(1,), low=low, high=high)
 
-        return random_weight * tf.keras.losses.categorical_crossentropy(
-            self.input, self.decoder
-        )
+        return random_weight * tf.keras.losses.categorical_crossentropy(self.input, self.decoder)
 
 
 class CharBasedTextFeature(BaseFeature):
     """
     A class to process the text features
     """
+
     def __init__(
         self,
         name: str,
@@ -380,8 +386,9 @@ class CharBasedTextFeature(BaseFeature):
         self.rnn_units = rnn_units
         self.rnn_unit = LSTM
         self.dropout = dropout
+        self.feature_type = "text"
 
-    def fit(self, data: pd.DataFrame):
+    def fit(self, data: pd.DataFrame, **kwargs):
         from tensorflow.keras.preprocessing.text import Tokenizer
 
         if len(data.columns) > 1:
@@ -397,7 +404,6 @@ class CharBasedTextFeature(BaseFeature):
         self.tokenizer = tokenizer
 
     def transform(self, data: pd.DataFrame) -> np.ndarray:
-
         if len(data.columns) > 1:
             raise Exception("CharBasedTextFeature can work only with one text column")
 
@@ -416,113 +422,96 @@ class CharBasedTextFeature(BaseFeature):
         # return data_gen
         return K.one_hot(K.cast(data_gen, "int32"), self.vocab_size)
 
-    def top_k_top_p_filtering(
-        self,
-        logits: np.ndarray,
-        top_k: int = 0,
-        top_p: float = 0,
-        filter_value: int = -1e8,
+    @staticmethod
+    def _top_p_filtering(
+            logits: np.ndarray,
+            top_p: float = 0.9
     ):
-        """Filter a distribution of logits using top-k and/or nucleus (top-p) filtering
-        https://gist.github.com/thomwolf/1a5a29f6962089e871b94cbd09daf317
-        https://github.com/ari-holtzman/degen
-        Args:
-            logits: logits distribution shape (..., vocabulary size)
-            top_k: >0 keep only top k tokens with highest probability (top-k filtering).
-            top_p: >0.0 keep the top tokens with cumulative probability >= top_p (nucleus filtering).
-            filter_value: value to replace small values with.
-        """
+        # Convert logits to TensorFlow tensor
+        logits = tf.convert_to_tensor(logits, dtype=tf.float32)
 
-        def custom_scatter_(zeros, index, src):
-            """Custom implementation of torch.Tensor.scatter_()
-            args:
-                zeros: np.ndarray placeholder for logits shape like logits
-                index: np.ndarray sorted indeces for scatters
-                src: np.ndarray indeces to remove
-            """
-            # zeros[i][index[i][j]] = src[i][j]
-            for i in range(len(index)):
-                for j in range(len(index[i])):
-                    zeros[i][index[i][j]] = src[i][j]
+        # Sort logits and get sorted indices
+        sorted_logits = tf.sort(logits, direction="DESCENDING", axis=-1)
+        sorted_indices = tf.argsort(logits, direction="DESCENDING", axis=-1)
 
-            return zeros
+        # Calculate cumulative probabilities
+        cumulative_probs = tf.cumsum(sorted_logits, axis=-1)
 
-        logits_removed = logits
-        # top_k = min(top_k, logits.size(-1)) # Safety check
-        if top_k > 0:
-            # Remove all tokens with a probability less than the last token of the top-k
-            indices_to_remove = logits < tf.math.top_k(logits, top_k)[0][..., -1, None]
-            logits[indices_to_remove] = filter_value
+        # Remove tokens with cumulative probability above the threshold
+        sorted_indices_to_remove = cumulative_probs >= top_p
 
-        if top_p > 0.0:
-            sorted_logits = tf.sort(logits, direction="DESCENDING")
-            sorted_indices = tf.argsort(logits, direction="DESCENDING")
-            cumulative_probs = tf.cumsum(tf.nn.softmax(sorted_logits, axis=-1), axis=-1)
+        # Shift the indices to the right to keep also the first token above the threshold
+        zeros_for_shift = tf.zeros_like(sorted_indices_to_remove[:, :, :1], dtype=tf.bool)
+        sorted_indices_to_remove = tf.concat(
+            [zeros_for_shift, sorted_indices_to_remove[:, :, :-1]],
+            axis=-1
+        )
 
-            # Remove tokens with cumulative probability above the threshold
-            sorted_indices_to_remove = cumulative_probs >= top_p
+        # Create a mask for indices to remove
+        batch_size, seq_length, vocab_size = logits.shape
 
-            # Shift the indices to the right to keep also the first token above the threshold
-            zeros_for_shift = tf.zeros(
-                [tf.shape(sorted_indices_to_remove)[0], 1], dtype=tf.bool
-            )
-            sorted_indices_to_remove = tf.concat(
-                (zeros_for_shift, sorted_indices_to_remove[:, :-1]), axis=1
-            )
+        batch_indices = tf.repeat(tf.range(batch_size), seq_length * vocab_size)
+        feature_length_indices = tf.tile(
+            tf.repeat(tf.range(seq_length), vocab_size),
+            [batch_size]
+        )
+        vocab_selection_indices = tf.reshape(sorted_indices, [-1])
+        update_indices = tf.stack(
+            [batch_indices, feature_length_indices, vocab_selection_indices],
+            axis=1
+        )
+        flattened_update_values = tf.reshape(sorted_indices_to_remove, [-1])
+        indices_to_remove = tf.tensor_scatter_nd_update(
+            tf.zeros_like(logits, dtype=sorted_indices_to_remove.dtype),
+            update_indices,
+            flattened_update_values,
+        )
 
-            # sorted_indices - ids of columns to replace
-            # line in sorted_indices corresponds to line in sorted_indices_to_remove
+        # Apply the filter value to the logits
+        logits_removed = tf.where(indices_to_remove, tf.fill(indices_to_remove.shape, 0.0), logits)
 
-            # indices are row and column coordinates like [[0, 5], ... [0, 2], [1, 7], ..., [1, 6], ...]
-            row_numbers = tf.reshape(
-                tf.repeat(tf.range(sorted_indices.shape[0]), sorted_indices.shape[1]),
-                [sorted_indices.shape[0] * sorted_indices.shape[1]],
-            )
-            flattened_sorted_indices = tf.reshape(
-                sorted_indices, [sorted_indices.shape[0] * sorted_indices.shape[1]]
-            )
-            update_indices = tf.stack([row_numbers, flattened_sorted_indices], axis=1)
-            # update values correspond to indices [False, False, True, .... True]
-            flattened_update_values = tf.reshape(sorted_indices_to_remove, [-1])
-            # replace old index values with new update values
-            indices_to_remove = tf.tensor_scatter_nd_update(
-                tf.zeros_like(logits, dtype=sorted_indices_to_remove.dtype),
-                update_indices,
-                flattened_update_values,
-            )
-            # fill indices with one filter_value
-            logits_removed = tf.where(
-                indices_to_remove,
-                tf.fill(indices_to_remove.shape, filter_value),
-                logits,
-            )
+        return logits_removed.numpy().astype(np.float64)
 
-        return logits_removed
+    @staticmethod
+    def _top_k_filtering(
+            logits: np.ndarray,
+            top_k: int = 0
+    ):
+        indices_to_remove = logits < tf.math.top_k(logits, top_k)[0][..., -1, None]
+        logits[indices_to_remove] = 0.0
+        return logits
+
+    def _process_batch(self, batch: np.ndarray) -> List[str]:
+        probs = tf.nn.softmax(batch, axis=-1).numpy().astype(float)
+        probs = self._top_p_filtering(probs, top_p=0.9)
+        # probs = self._top_k_filtering(probs, top_k=6)
+        # TODO: select top_k based on inverse_dict length
+
+        probs /= probs.sum(axis=2, keepdims=True)
+
+        multinomial_samples = np.apply_along_axis(
+            lambda x: np.argmax(np.random.multinomial(1, x)), -1, probs
+        )
+
+        chars_array = np.vectorize(lambda x: self.tokenizer.inverse_dict.get(x, ''))(
+            multinomial_samples)
+
+        # Convert tokens to words
+        words = ["".join(sample) for sample in chars_array]
+
+        return words
 
     def inverse_transform(self, data: np.ndarray, **kwargs) -> List[str]:
-        top_p = 0.9
-        if len(kwargs) > 0:
-            top_p = kwargs["top_p"]
 
-        out = []
-        for batch in data:
-            # batch shape (self.text_max_len, self.vocab_size)
-            logits = self.top_k_top_p_filtering(
-                tf.convert_to_tensor(batch), top_p=top_p
-            )
-            probs = tf.nn.softmax(logits, axis=-1).numpy().astype(float)
-            probs /= probs.sum(axis=1)[:, None]
-            multinomial_samples = np.apply_along_axis(
-                lambda x: np.argmax(np.random.multinomial(1, x)), 1, probs
-            )
-            tokens = list(multinomial_samples)
-            word = "".join(
-                self.tokenizer.inverse_dict[x]
-                for x in tokens
-                if x in self.tokenizer.inverse_dict.keys()
-            )
-            out.append(word)
-        return out
+        batch_size = 10000
+        num_batches = (len(data) + batch_size - 1) // batch_size
+        feature_values = []
+
+        for i in range(num_batches):
+            batch = data[i * batch_size: (i + 1) * batch_size]
+            feature_values.extend(self._process_batch(batch))
+
+        return feature_values
 
     @lazy
     def input(self) -> tf.Tensor:
@@ -534,9 +523,7 @@ class CharBasedTextFeature(BaseFeature):
 
     @lazy
     def encoder(self) -> tf.Tensor:
-        rnn_encoder_layer = Bidirectional(
-            self.rnn_unit(self.rnn_units, return_sequences=False)
-        )
+        rnn_encoder_layer = Bidirectional(self.rnn_unit(self.rnn_units, return_sequences=False))
 
         rnn_econder = rnn_encoder_layer(self.input)
         return rnn_econder
@@ -547,9 +534,7 @@ class CharBasedTextFeature(BaseFeature):
 
         decoder_layers.append(RepeatVector(self.text_max_len))
         decoder_layers.append(self.rnn_unit(self.rnn_units, return_sequences=True))
-        decoder_layers.append(
-            TimeDistributed(Dense(self.vocab_size, activation="linear"))
-        )
+        decoder_layers.append(TimeDistributed(Dense(self.vocab_size, activation="linear")))
 
         return decoder_layers
 
@@ -572,24 +557,60 @@ class CharBasedTextFeature(BaseFeature):
             Exception("Decoder isn't created")
 
         return self.weight * K.mean(
-            tf.nn.softmax_cross_entropy_with_logits(
+            tf.compat.v1.nn.softmax_cross_entropy_with_logits_v2(
                 labels=self.input, logits=self.decoder
             )
         )
 
 
+class EmailFeature(CharBasedTextFeature):
+    def __init__(
+        self,
+        name: str,
+        text_max_len: int,
+        rnn_units: int = 128,
+        dropout: int = 0,
+        domain: str = 'tdspora.ai'
+    ):
+        super().__init__(name=name,
+                         text_max_len=text_max_len,
+                         rnn_units=rnn_units,
+                         dropout=dropout)
+        self.domain = domain
+
+    def fit(self, data: pd.DataFrame, **kwargs):
+        super().fit(self.extract_email_name(data))
+
+    @staticmethod
+    def extract_email_name(data: pd.DataFrame) -> pd.DataFrame:
+        pattern = r'^([^@]+).*$'  # returns all the string if there's no "@" in it
+        return data.iloc[:, 0].str.extract(pattern).rename({0: data.columns[0]}, axis=1).fillna('')
+
+    def transform(self, data: pd.DataFrame) -> np.ndarray:
+        if len(data.columns) > 1:
+            raise Exception("CharBasedTextFeature can work only with one text column")
+
+        # TODO: add logic for emails to Dataset._preprocess_nan_cols
+        return super().transform(self.extract_email_name(data))
+
+    def inverse_transform(self, data: np.ndarray, **kwargs) -> List[str]:
+        return [
+            s + "@" +
+            self.domain for s in super().inverse_transform(data, **kwargs)
+        ]
+
+
 class DateFeature(BaseFeature):
     """
     A class to process datetime features
     """
-    def __init__(self, name, decoder_layers=(60,), weight_randomizer=None):
 
+    def __init__(self, name, decoder_layers=(60,), weight_randomizer=None):
         if decoder_layers is None:
             decoder_layers = ()
         elif isinstance(decoder_layers, int):
             decoder_layers = (decoder_layers,)
 
-        # TODO: teset features
         if weight_randomizer is None or (
             isinstance(weight_randomizer, bool) and not weight_randomizer
         ):
@@ -602,82 +623,43 @@ class DateFeature(BaseFeature):
         super().__init__(name=name)
         self.decoder_layers = decoder_layers
         self.weight_randomizer = weight_randomizer
+        self.feature_type = "numeric"
 
-    @staticmethod
-    def __validate_format(date_text: pd.DataFrame):
-        """
-        Define the most common date format.
-        Supported date formats -
-        MM/DD/YYYY; MM-DD-YYYY; DD/MM/YYYY; DD-MM-YYYY;
-        YYYY/MM/DD; YYYY-MM-DD; MMM DD, YYYY; MMM DD YYYY;
-        DD MMM YYYY; YYYY-MM-DD HH:MM:SS
-
-        Not supported formats -
-        MM/DD/YY, DD/MM/YY, YY/MM/DD, MM-DD-YY, DD-MM-YY
-
-        """
-        pattern = r"\s{0,1}\d+[-/\\:]\s{0,1}\d+[-/\\:]\s{0,1}\d+|" \
-                  r"[A-Z][a-z]+ \d{1,2} \d{4}|" \
-                  r"[A-Z][a-z]+ \d{1,2}, \d{4}|" \
-                  r"\d{2} [A-Z][a-z]+ \d{4}"
-        types = []
-        sample = date_text.dropna().sample(100, replace=len(date_text) <= 100).values
-        for i in sample:
-            date_format = guess_datetime_format(re.match(pattern, i[0]).group(0))
-            types.append(date_format)
-        if not list(filter(lambda x: bool(x), types)) or not types:
-            return "%d-%m-%Y"
-        if Counter(types).most_common(1)[0][0] is None:
-            return Counter(types).most_common(2)[1][0]
-        return Counter(types).most_common(1)[0][0]
-
-    def fit(self, data):
-        self.date_format = self.__validate_format(data)
-        data = chain.from_iterable(data.values)
-        data = pd.DataFrame(list(map(lambda d: pd.Timestamp(d).value, data)))
-        self.is_positive = (data >= 0).sum().item() >= len(data) * 0.99
-        normality = shapiro(data.sample(n=min(len(data), 500))).pvalue
-        data = np.array(data).reshape(-1, 1)
+    def fit(self, data, **kwargs):
+        self.date_format = kwargs["date_mapping"][self.original_name]
+        self.data = chain.from_iterable(data.values)
+        self.data = pd.DataFrame(
+            list(
+                map(
+                    lambda d: datetime_to_timestamp(d, self.date_format),
+                    self.data
+                )
+            )
+        )
+        self.is_positive = (self.data >= 0).sum().item() >= len(self.data) * 0.99
+        normality = shapiro(self.data.sample(n=min(len(self.data), 500))).pvalue
+        self.data = np.array(self.data).reshape(-1, 1)
 
         self.scaler = StandardScaler() if normality >= 0.05 else MinMaxScaler()
-        self.scaler.fit(data)
-        self.input_dimension = data.shape[1]
+        self.scaler.fit(self.data)
+        self.input_dimension = self.data.shape[1]
 
     def transform(self, data):
-        data = chain.from_iterable(data.values)
-        data = list(map(lambda d: pd.Timestamp(d).value, data))
-        data = np.array(data).reshape(-1, 1)
-        return self.scaler.transform(data)
+        return self.scaler.transform(self.data)
 
     def inverse_transform(self, data):
-        max_allowed_time_ns = int(9.2E18)
-        min_allowed_time_ns = int(-9.2E18)
         unscaled = self.scaler.inverse_transform(data)
         unscaled = chain.from_iterable(unscaled)
-        if self.is_positive:
-            return list(
-                map(
-                    lambda l: pd.Timestamp(
-                        abs(min(max_allowed_time_ns, int(l)))
-                    ).strftime(self.date_format),
-                    unscaled,
-                )
-            )
-        else:
-            return list(
-                map(
-                    lambda l: pd.Timestamp(
-                        max(min(max_allowed_time_ns, int(l)), min_allowed_time_ns)
-                    ).strftime(self.date_format),
-                    unscaled,
-                )
+        return list(
+            map(
+                lambda t: timestamp_to_datetime(int(t)).strftime(self.date_format),
+                unscaled,
             )
+        )
 
     @lazy
     def input(self):
-        return Input(
-            shape=(self.input_dimension,), name="input_%s" % self.name, dtype="float64"
-        )
+        return Input(shape=(self.input_dimension,), name="input_%s" % self.name, dtype="float64")
 
     @lazy
     def encoder(self):
